import math
import numpy as np

from wx import glcanvas
import OpenGL
from OpenGL.GL import *
from OpenGL.GL.ARB.shader_objects import *
from OpenGL.GL.ARB.vertex_shader import *
#from OpenGL.GL.ARB.vertex_array_object import *
from OpenGL.GL.ARB.draw_instanced import *
from OpenGL.GL.ARB.instanced_arrays import *
from OpenGL.GL import shaders

from ctypes import sizeof, c_float, c_void_p, c_uint

import MatrixUtil as matrix

CUBE_OFFS = [
    (1,1,1),(0,1,1),(0,0,1),(1,0,1),
    (0,0,0),(0,1,0),(1,1,0),(1,0,0),
    (1,1,1),(1,1,0),(0,1,0),(0,1,1),
    (0,0,0),(1,0,0),(1,0,1),(0,0,1),
    (1,1,1),(1,0,1),(1,0,0),(1,1,0),
    (0,0,0),(0,0,1),(0,1,1),(0,1,0),
]

# vertices to be used by OpenGL
# first three components are position
# second three components are normal
CUBE_VERTICES = np.array([
    # top
    0, 0, 1, 0, 0, 1,
    0, 1, 1, 0, 0, 1,
    1, 0, 1, 0, 0, 1,
    1, 1, 1, 0, 0, 1,
    
    # bottom
    0, 0, 0, 0, 0,-1,
    0, 1, 0, 0, 0,-1,
    1, 0, 0, 0, 0,-1,
    1, 1, 0, 0, 0,-1,

    # left
    0, 0, 0,-1, 0, 0,
    0, 0, 1,-1, 0, 0,
    0, 1, 0,-1, 0, 0,
    0, 1, 1,-1, 0, 0,

    # right
    1, 0, 0, 1, 0, 0,
    1, 0, 1, 1, 0, 0,
    1, 1, 0, 1, 0, 0,
    1, 1, 1, 1, 0, 0,

    # front
    0, 0, 0, 0,-1, 0,
    0, 0, 1, 0,-1, 0,
    1, 0, 0, 0,-1, 0,
    1, 0, 1, 0,-1, 0,

    # back
    0, 1, 0, 0, 1, 0,
    0, 1, 1, 0, 1, 0,
    1, 1, 0, 0, 1, 0,
    1, 1, 1, 0, 1, 0,
], dtype=np.float32)

CUBE_INDICES = np.array([
    # top
    0, 2, 1, 3, 1, 2,
    # bottom
    4, 5, 6, 7, 6, 5,
    # left
    8, 9, 10, 11, 10, 9,
    # right
    12, 14, 13, 15, 13, 14,
    # front
    16, 18, 17, 19, 17, 18,
    # back
    20, 21, 22, 23, 22, 21,
], dtype=np.uint32)

OUTLINE_VERTICES = np.array([
    0, 0, 0, 1, 0, 0,
    0, 0, 0, 0, 1, 0,
    0, 0, 0, 0, 0, 1,
    1, 1, 1, 1, 1, 0,
    1, 1, 1, 1, 0, 1,
    1, 1, 1, 0, 1, 1,
    0, 0, 1, 0, 1, 1,
    0, 0, 1, 1, 0, 1,
    1, 1, 0, 0, 1, 0,
    1, 1, 0, 1, 0, 0,
    0, 1, 0, 0, 1, 1,
    1, 0, 0, 1, 0, 1,
], dtype=np.float32)

COMPASS_VERTICES = np.array([
    0, 0, 0, 1, 0, 0, 1, 0, 0, 1, 0, 0,
    0, 0, 0, 0, 1, 0, 0, 1, 0, 0, 1, 0,
    0, 0, 0, 0, 0, 1, 0, 0, 1, 0, 0, 1,
], dtype=np.float32)

QUAD_VERTICES = np.array([
    0, 0, 0,
    1, 0, 0,
    1, 1, 0,
    1, 1, 0,
    0, 1, 0,
    0, 0, 0,
], dtype=np.float32)

# constants for OpenGL buffers
INT_SIZE   = sizeof(c_int)
UINT_SIZE  = sizeof(c_uint)
FLOAT_SIZE = sizeof(c_float)

POSITION_OFFSET = c_void_p(0 * FLOAT_SIZE)
NORMAL_OFFSET   = c_void_p(3 * FLOAT_SIZE)
OFFSET_OFFSET   = c_void_p(0 * INT_SIZE)
COLOR_OFFSET    = c_void_p(3 * INT_SIZE)
ID_OFFSET       = c_void_p(6 * INT_SIZE)
VERTEX_SIZE     = 3 * FLOAT_SIZE

# OpenGL shader attribute locations
POSITION_LOC = 0
NORMAL_LOC   = 1
OFFSET_LOC   = 2
COLOR_LOC    = 3
ID_LOC       = 4

gl_major_version = 0
gl_minor_version = 0
gl_33 = False


# Since render to texture is being used to assign pixels to tiles for the mouseover feature,
# We need to re-render the screen texture to the actual screen instead of just a texture
# The ScreenQuad is the textured quad to which we render the screen texture
class ScreenQuad:

    def __init__(self):
        pass

    def initGraphics(self):
        self.vertexArray = glGenVertexArrays(1)
        self.vertexBuffer = glGenBuffers(1)

        glBindVertexArray(self.vertexArray)

        glBindBuffer(GL_ARRAY_BUFFER, self.vertexBuffer)
        glBufferData(GL_ARRAY_BUFFER, len(QUAD_VERTICES) * FLOAT_SIZE, QUAD_VERTICES, GL_STATIC_DRAW)

        offset = c_void_p(0 * FLOAT_SIZE)
        glEnableVertexAttribArray(POSITION_LOC)
        glVertexAttribPointer(POSITION_LOC, 3, GL_FLOAT, GL_FALSE, 0, offset)

        glBindVertexArray(0)

    def Draw(self):
        glBindVertexArray(self.vertexArray)
        glDrawArrays(GL_TRIANGLES, 0, 18)
        glBindVertexArray(0)

# The compass is rendered in the bottom left corner of the screen to show the current orientation of the camera
class Compass:

    def __init__(self):
        pass

    def initGraphics(self):
        self.vertexArray = glGenVertexArrays(1)
        self.vertexBuffer = glGenBuffers(1)

        glBindVertexArray(self.vertexArray)

        glBindBuffer(GL_ARRAY_BUFFER, self.vertexBuffer)
        glBufferData(GL_ARRAY_BUFFER, len(COMPASS_VERTICES) * FLOAT_SIZE, COMPASS_VERTICES, GL_STATIC_DRAW)

        offset = c_void_p(0 * FLOAT_SIZE)
        glEnableVertexAttribArray(POSITION_LOC)
        glVertexAttribPointer(POSITION_LOC, 3, GL_FLOAT, GL_FALSE, 2 * VERTEX_SIZE, offset)
        offset = c_void_p(3 * FLOAT_SIZE)
        glEnableVertexAttribArray(COLOR_LOC)
        glVertexAttribPointer(COLOR_LOC, 3, GL_FLOAT, GL_FALSE, 2 * VERTEX_SIZE, offset)

        glBindVertexArray(0)

    def Draw(self):
        glLineWidth(2.0)
        glBindVertexArray(self.vertexArray)
        glDrawArrays(GL_LINES, 0, 6)
        glBindVertexArray(0)

class Buffer:
    # 3 ints for color, 3 ints for offset
    TILE_DATA_SIZE = 3 * INT_SIZE + 3 * INT_SIZE

    def __init__(self, capacity=2000000):
        self.capacity = capacity
        self.size = 0
        self.updateTiles = []
        self.modified = False

    def initGraphics(self):
        self.vertexArray    = glGenVertexArrays(1)
        self.cubeBuffer     = glGenBuffers(1) # Vertex buffer to contain cube vertex info
        self.indexBuffer    = glGenBuffers(1) # Vertex buffer to contain cube index info
        self.instanceBuffer = glGenBuffers(1) # Vertex buffer to contain instancing info

        self.lineArray      = glGenVertexArrays(1)
        self.lineBuffer     = glGenBuffers(1)

        glBindVertexArray(self.vertexArray)

        glBindBuffer(GL_ARRAY_BUFFER, self.cubeBuffer)
        glBufferData(GL_ARRAY_BUFFER, len(CUBE_VERTICES) * FLOAT_SIZE, CUBE_VERTICES, GL_STATIC_DRAW)
        glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, self.indexBuffer)
        glBufferData(GL_ELEMENT_ARRAY_BUFFER, len(CUBE_INDICES) * UINT_SIZE, CUBE_INDICES, GL_STATIC_DRAW)

        glEnableVertexAttribArray(POSITION_LOC)
        glVertexAttribPointer(POSITION_LOC, 3, GL_FLOAT, GL_FALSE, 2 * VERTEX_SIZE, POSITION_OFFSET)
        glEnableVertexAttribArray(NORMAL_LOC)
        glVertexAttribPointer(NORMAL_LOC, 3, GL_FLOAT, GL_FALSE, 2 * VERTEX_SIZE, NORMAL_OFFSET)

        glBindBuffer(GL_ARRAY_BUFFER, self.instanceBuffer)
        glBufferData(GL_ARRAY_BUFFER, self.capacity * Buffer.TILE_DATA_SIZE, None, GL_DYNAMIC_DRAW)

        glEnableVertexAttribArray(OFFSET_LOC)
        glVertexAttribIPointer(OFFSET_LOC, 3, GL_INT, Buffer.TILE_DATA_SIZE, OFFSET_OFFSET)
        
        glEnableVertexAttribArray(COLOR_LOC)
        glVertexAttribIPointer(COLOR_LOC, 3, GL_INT, Buffer.TILE_DATA_SIZE, COLOR_OFFSET)

        if gl_33:
            glVertexAttribDivisor(POSITION_LOC, 0)
            glVertexAttribDivisor(NORMAL_LOC, 0)
            glVertexAttribDivisor(OFFSET_LOC, 1)
            glVertexAttribDivisor(COLOR_LOC, 1)
        else:
            glVertexAttribDivisorARB(POSITION_LOC, 0)
            glVertexAttribDivisorARB(NORMAL_LOC, 0)
            glVertexAttribDivisorARB(OFFSET_LOC, 1)
            glVertexAttribDivisorARB(COLOR_LOC, 1)

        # Array to render outline of cube
        glBindVertexArray(self.lineArray)

        glBindBuffer(GL_ARRAY_BUFFER, self.lineBuffer)
        glBufferData(GL_ARRAY_BUFFER, len(OUTLINE_VERTICES) * FLOAT_SIZE, OUTLINE_VERTICES, GL_STATIC_DRAW)
        glEnableVertexAttribArray(POSITION_LOC)
        glVertexAttribPointer(POSITION_LOC, 3, GL_FLOAT, GL_FALSE, VERTEX_SIZE, POSITION_OFFSET)

        glBindBuffer(GL_ARRAY_BUFFER, self.instanceBuffer)
        glBufferData(GL_ARRAY_BUFFER, self.capacity * Buffer.TILE_DATA_SIZE, None, GL_DYNAMIC_DRAW)

        glEnableVertexAttribArray(OFFSET_LOC)
        glVertexAttribIPointer(OFFSET_LOC, 3, GL_INT, Buffer.TILE_DATA_SIZE, OFFSET_OFFSET)
        
        glEnableVertexAttribArray(COLOR_LOC)
        glVertexAttribIPointer(COLOR_LOC, 3, GL_INT, Buffer.TILE_DATA_SIZE, COLOR_OFFSET)

        if gl_33:
            glVertexAttribDivisor(POSITION_LOC, 0)
            glVertexAttribDivisor(OFFSET_LOC, 1)
            glVertexAttribDivisor(COLOR_LOC, 1)
        else:
            glVertexAttribDivisorARB(POSITION_LOC, 0)
            glVertexAttribDivisorARB(OFFSET_LOC, 1)
            glVertexAttribDivisorARB(COLOR_LOC, 1)

        glBindVertexArray(0)

    def deleteBuffer(self):
        #TODO: delete OpenGL variables here
        self.updateTiles = []
        self.size = 0

    def isFull(self):
        return self.size + len(self.updateTiles) >= self.capacity

    def AddTile(self, tile):
        if self.size + len(self.updateTiles) < self.capacity:
            self.updateTiles.append(tile)
            self.modified = True
            return True
        else:
            return False

    def UpdateBuffer(self):
        if self.modified:
            data = np.empty(len(self.updateTiles)*6, dtype=np.int32)
            i = 0
            for tile in self.updateTiles:
                offset = np.array(tile.coords, dtype=np.int32)
                color = np.array(tile.type.colorVector, dtype=np.int32)
                for j in range(3):
                    data[i] = offset[j]
                    i += 1
                for j in range(3):
                    data[i] = color[j]
                    i += 1
            glBindVertexArray(self.vertexArray)
            glBindBuffer(GL_ARRAY_BUFFER, self.instanceBuffer)
            glBufferSubData(GL_ARRAY_BUFFER, self.size * Buffer.TILE_DATA_SIZE, len(self.updateTiles) * Buffer.TILE_DATA_SIZE, data)

            glBindVertexArray(self.lineArray)
            glBindBuffer(GL_ARRAY_BUFFER, self.instanceBuffer)
            glBufferSubData(GL_ARRAY_BUFFER, self.size * Buffer.TILE_DATA_SIZE, len(self.updateTiles) * Buffer.TILE_DATA_SIZE, data)
            glBindVertexArray(0)

            self.size += len(self.updateTiles)
            self.modified = False

    def DrawTiles(self):
        if self.modified:
            self.UpdateBuffer()

        glBindVertexArray(self.vertexArray)
        if gl_33:
           glDrawElementsInstanced(GL_TRIANGLES, 36, GL_UNSIGNED_INT, None, self.size)
        else:
           glDrawElementsInstancedARB(GL_TRIANGLES, 36, GL_UNSIGNED_INT, None, self.size)
        glBindVertexArray(0)

    def DrawOutlines(self, lineWidth=1.0):
        if self.modified:
            self.UpdateBuffers()

        glLineWidth(lineWidth)

        glBindVertexArray(self.lineArray)
        if gl_33:
            glDrawArraysInstanced(GL_LINES, 0, 24, self.size)
        else:
            glDrawArraysInstancedARB(GL_LINES, 0, 24, self.size)
        glBindVertexArray(0)

# A Spacial Partition that holds tiles to be rendered
# This allows for smaller buffers to store the groups of tiles which makes instancing faster
# These take up blocks of SIZE^3 tiles 
class Partition:
    SIZE = 64
    DIMENSIONS = [SIZE] * 3

    # 3 ints for color, 3 ints for offset
    TILE_DATA_SIZE = 3 * INT_SIZE + 3 * INT_SIZE

    def __init__(self, offset):
        self.offset = [x * Partition.SIZE for x in offset]
        self.center = np.array([x + Partition.SIZE / 2.0 for x in self.offset], dtype=float)

        self.modified = False
        self.tiles = {}

    def initGraphics(self):
        self.vertexArray    = glGenVertexArrays(1)
        self.cubeBuffer     = glGenBuffers(1) # Vertex buffer to contain cube vertex info
        self.indexBuffer    = glGenBuffers(1) # Vertex buffer to contain cube index info
        self.instanceBuffer = glGenBuffers(1) # Vertex buffer to contain instancing info

        self.lineArray      = glGenVertexArrays(1)
        self.lineBuffer     = glGenBuffers(1)

        glBindVertexArray(self.vertexArray)

        glBindBuffer(GL_ARRAY_BUFFER, self.cubeBuffer)
        glBufferData(GL_ARRAY_BUFFER, len(CUBE_VERTICES) * FLOAT_SIZE, CUBE_VERTICES, GL_STATIC_DRAW)

        glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, self.indexBuffer)
        glBufferData(GL_ELEMENT_ARRAY_BUFFER, len(CUBE_INDICES) * UINT_SIZE, CUBE_INDICES, GL_STATIC_DRAW)

        glEnableVertexAttribArray(POSITION_LOC)
        glVertexAttribPointer(POSITION_LOC, 3, GL_FLOAT, GL_FALSE, 2 * VERTEX_SIZE, POSITION_OFFSET)
        glEnableVertexAttribArray(NORMAL_LOC)
        glVertexAttribPointer(NORMAL_LOC, 3, GL_FLOAT, GL_FALSE, 2 * VERTEX_SIZE, NORMAL_OFFSET)

        data = np.zeros(Partition.TILE_DATA_SIZE * 1, dtype=np.int32)

        glBindBuffer(GL_ARRAY_BUFFER, self.instanceBuffer)
        glBufferData(GL_ARRAY_BUFFER, 1 * Partition.TILE_DATA_SIZE, data, GL_STATIC_DRAW)

        glEnableVertexAttribArray(OFFSET_LOC)
        glVertexAttribIPointer(OFFSET_LOC, 3, GL_INT, Partition.TILE_DATA_SIZE, OFFSET_OFFSET)
        
        glEnableVertexAttribArray(COLOR_LOC)
        glVertexAttribIPointer(COLOR_LOC, 3, GL_INT, Partition.TILE_DATA_SIZE, COLOR_OFFSET)

        if gl_33:
            glVertexAttribDivisor(POSITION_LOC, 0)
            glVertexAttribDivisor(NORMAL_LOC, 0)
            glVertexAttribDivisor(OFFSET_LOC, 1)
            glVertexAttribDivisor(COLOR_LOC, 1)
        else:
            glVertexAttribDivisorARB(POSITION_LOC, 0)
            glVertexAttribDivisorARB(NORMAL_LOC, 0)
            glVertexAttribDivisorARB(OFFSET_LOC, 1)
            glVertexAttribDivisorARB(COLOR_LOC, 1)

        # Array to render outline of cube
        glBindVertexArray(self.lineArray)

        glBindBuffer(GL_ARRAY_BUFFER, self.lineBuffer)
        glBufferData(GL_ARRAY_BUFFER, len(OUTLINE_VERTICES) * FLOAT_SIZE, OUTLINE_VERTICES, GL_STATIC_DRAW)

        glEnableVertexAttribArray(POSITION_LOC)
        glVertexAttribPointer(POSITION_LOC, 3, GL_FLOAT, GL_FALSE, VERTEX_SIZE, POSITION_OFFSET)

        glBindBuffer(GL_ARRAY_BUFFER, self.instanceBuffer)
        # glBufferData(GL_ARRAY_BUFFER, 1 * Partition.TILE_DATA_SIZE, data, GL_STATIC_DRAW)

        glEnableVertexAttribArray(OFFSET_LOC)
        glVertexAttribIPointer(OFFSET_LOC, 3, GL_INT, Partition.TILE_DATA_SIZE, OFFSET_OFFSET)
        
        glEnableVertexAttribArray(COLOR_LOC)
        glVertexAttribIPointer(COLOR_LOC, 3, GL_INT, Partition.TILE_DATA_SIZE, COLOR_OFFSET)

        if gl_33:
            glVertexAttribDivisor(POSITION_LOC, 0)
            glVertexAttribDivisor(OFFSET_LOC, 1)
            glVertexAttribDivisor(COLOR_LOC, 1)
        else:
            glVertexAttribDivisorARB(POSITION_LOC, 0)
            glVertexAttribDivisorARB(OFFSET_LOC, 1)
            glVertexAttribDivisorARB(COLOR_LOC, 1)

        glBindVertexArray(0)

        del data

    def deletePartition(self):
        #TODO: delete OpenGL variables here
        del self.tiles

    def AddTile(self, tile):
        pos = tuple([x - y for x, y in zip(tile.coords, self.offset)])

        self.tiles[pos] = tile
        self.modified = True
        return True

    def RemoveTile(self, position):
        pos = tuple([x - y for x, y in zip(position, self.offset)])

        if pos in self.tiles:
            del self.tiles[pos]
            self.modified = True
        return True

    def UpdateBuffers(self):
        if self.modified:
            data = np.array([], dtype=np.int32)
            for tile in self.tiles.itervalues():
                offset = np.array(tile.coords, dtype=np.int32)
                color = np.array(tile.type.colorVector, dtype=np.int32)
                data = np.append(data, offset)
                data = np.append(data, color)
            glBindVertexArray(self.vertexArray)
            glBindBuffer(GL_ARRAY_BUFFER, self.instanceBuffer)
            glBufferData(GL_ARRAY_BUFFER, len(self.tiles) * Partition.TILE_DATA_SIZE, data, GL_STATIC_DRAW)

            glEnableVertexAttribArray(OFFSET_LOC)
            glVertexAttribIPointer(OFFSET_LOC, 3, GL_INT, Partition.TILE_DATA_SIZE, OFFSET_OFFSET)
            
            glEnableVertexAttribArray(COLOR_LOC)
            glVertexAttribIPointer(COLOR_LOC, 3, GL_INT, Partition.TILE_DATA_SIZE, COLOR_OFFSET)

            glBindVertexArray(self.lineArray)

            glBindBuffer(GL_ARRAY_BUFFER, self.instanceBuffer)
            # glBufferData(GL_ARRAY_BUFFER, 1 * Partition.TILE_DATA_SIZE, data, GL_STATIC_DRAW)

            glEnableVertexAttribArray(OFFSET_LOC)
            glVertexAttribIPointer(OFFSET_LOC, 3, GL_INT, Partition.TILE_DATA_SIZE, OFFSET_OFFSET)
            
            glEnableVertexAttribArray(COLOR_LOC)
            glVertexAttribIPointer(COLOR_LOC, 3, GL_INT, Partition.TILE_DATA_SIZE, COLOR_OFFSET)

            glBindVertexArray(0)

            self.modified = False

    def DrawTiles(self):
        if self.modified:
            self.UpdateBuffers()

        glBindVertexArray(self.vertexArray)
        if gl_33:
           glDrawElementsInstanced(GL_TRIANGLES, 36, GL_UNSIGNED_INT, None, len(self.tiles))
        else:
           glDrawElementsInstancedARB(GL_TRIANGLES, 36, GL_UNSIGNED_INT, None, len(self.tiles))
        glBindVertexArray(0)

    def DrawOutlines(self, lineWidth=1.0):
        if self.modified:
            self.UpdateBuffers()

        glLineWidth(lineWidth)

        glBindVertexArray(self.lineArray)
        if gl_33:
            glDrawArraysInstanced(GL_LINES, 0, 24, len(self.tiles))
        else:
            glDrawArraysInstancedARB(GL_LINES, 0, 24, len(self.tiles))
        glBindVertexArray(0)


# A wrapper for OpenGL functions so that PyTAS3DCanvas doesn't have to be littered with OpenGL calls
class GraphicsWrapper:
    def __init__(self, size):
        self.size               = size
        self.drawOutlines       = True
        self.outlineWidth       = 1.0
        self.outlineDistance    = 1000.0

        self.tileBuffers        = []
        self.newBufferCount     = 0
        self.partitions         = {}
        self.newPartitionList   = []

        self.projectionMatrix = matrix.identity_matrix()
        self.rotationMatrix   = matrix.identity_matrix()
        self.viewMatrix       = matrix.identity_matrix()
        self.centerMatrix     = matrix.identity_matrix()
        self.invCenterMatrix  = matrix.identity_matrix()

        self.cubeShader    = None
        self.outlineShader = None
        self.compassShader = None
        self.screenShader  = None

        self.framebuffer = None
        self.depthbuffer = None
        self.screenTex   = None
        self.mouseTex    = None

        self.screenQuad = ScreenQuad()
        self.compass = Compass()

    def initShader(self, src):
        vs_src_file = open(src + '.vert')
        vs_src = vs_src_file.read()
        vs_src_file.close()
        vertexShader = shaders.compileShader(vs_src, GL_VERTEX_SHADER)

        fs_src_file = open(src + '.frag')
        fs_src = fs_src_file.read()
        fs_src_file.close()
        fragmentShader = shaders.compileShader(fs_src, GL_FRAGMENT_SHADER)

        shader = glCreateProgram()
        glAttachShader(shader, vertexShader)
        glAttachShader(shader, fragmentShader)

        glBindAttribLocation(shader, 0, "a_Position")
        glBindAttribLocation(shader, 1, "a_Normal")
        glBindAttribLocation(shader, 2, "a_TileOffset")
        glBindAttribLocation(shader, 3, "a_Color")

        glBindFragDataLocation(shader, 0, "f_Color")
        glBindFragDataLocation(shader, 1, "f_TileId")

        glLinkProgram(shader)

        glDeleteShader(vertexShader)
        glDeleteShader(fragmentShader)

        return shader

    def initTexture(self, width, height, int_tex = False):
        tex = glGenTextures(1)
        glBindTexture(GL_TEXTURE_2D, tex)
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
        glTexParameteri(GL_TEXTURE_2D, GL_GENERATE_MIPMAP, GL_FALSE)
        if int_tex:
            glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA32I, width, height, 0, GL_RGBA_INTEGER, GL_INT, None)
        else:
            glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB, width, height, 0, GL_RGBA, GL_UNSIGNED_BYTE, None)
        glBindTexture(GL_TEXTURE_2D, 0)
        return tex

    def resizeTexture(self, tex, width, height, int_tex = False):
        glBindTexture(GL_TEXTURE_2D, tex)
        if int_tex:
            glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA32I, width, height, 0, GL_RGBA_INTEGER, GL_INT, None)
        else:
            glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB, width, height, 0, GL_RGBA, GL_UNSIGNED_BYTE, None)
        glBindTexture(GL_TEXTURE_2D, 0)


    def initFramebuffer(self, width, height):
        self.screenTex = self.initTexture(width, height)
        self.mouseTex  = self.initTexture(width, height, True)

        self.depthbuffer = glGenRenderbuffers(1)
        glBindRenderbuffer(GL_RENDERBUFFER, self.depthbuffer)
        glRenderbufferStorage(GL_RENDERBUFFER, GL_DEPTH_COMPONENT, width, height)
        glBindRenderbuffer(GL_RENDERBUFFER, 0)

        self.framebuffer = glGenFramebuffers(1)
        glBindFramebuffer(GL_FRAMEBUFFER, self.framebuffer)
        glFramebufferRenderbuffer(GL_FRAMEBUFFER, GL_DEPTH_ATTACHMENT, GL_RENDERBUFFER, self.depthbuffer)
        glFramebufferTexture2D(GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT0, GL_TEXTURE_2D, self.screenTex, 0)
        glFramebufferTexture2D(GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT1, GL_TEXTURE_2D, self.mouseTex, 0)

        glDrawBuffers(2, [GL_COLOR_ATTACHMENT0, GL_COLOR_ATTACHMENT1])

        status = glCheckFramebufferStatus(GL_FRAMEBUFFER)
        if status != GL_FRAMEBUFFER_COMPLETE:
            PrintDebug("ERROR: Failed to create framebuffer!", ERROR_MSG)

        glBindFramebuffer(GL_FRAMEBUFFER, 0)

    def ResizeFramebuffer(self):
        if self.framebuffer is None:
            return

        width = self.size.width
        height = self.size.height

        self.resizeTexture(self.screenTex, width, height)
        self.resizeTexture(self.mouseTex, width, height, True)

        glBindRenderbuffer(GL_RENDERBUFFER, self.depthbuffer)
        glRenderbufferStorage(GL_RENDERBUFFER, GL_DEPTH_COMPONENT, width, height)
        glBindRenderbuffer(GL_RENDERBUFFER, 0)
        
        glBindFramebuffer(GL_FRAMEBUFFER, self.framebuffer)
        glFramebufferRenderbuffer(GL_FRAMEBUFFER, GL_DEPTH_ATTACHMENT, GL_RENDERBUFFER, self.depthbuffer)
        glFramebufferTexture2D(GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT0, GL_TEXTURE_2D, self.screenTex, 0)
        glFramebufferTexture2D(GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT1, GL_TEXTURE_2D, self.mouseTex, 0)

        glDrawBuffers(2, [GL_COLOR_ATTACHMENT0, GL_COLOR_ATTACHMENT1])

        status = glCheckFramebufferStatus(GL_FRAMEBUFFER)
        if status != GL_FRAMEBUFFER_COMPLETE:
            PrintDebug("ERROR: Failed to create framebuffer!", ERROR_MSG)

        glBindFramebuffer(GL_FRAMEBUFFER, 0)

    def InitGraphics(self):
        # GL initialization
        gl_major_version = glGetIntegerv(GL_MAJOR_VERSION)
        gl_minor_version = glGetIntegerv(GL_MINOR_VERSION)

        # Check if version is greater than 3.3
        gl_33 = (gl_major_version > 3) or (gl_major_version == 3 and gl_minor_version >= 3)

        print("GL VERSION: " + str(gl_major_version) + "." + str(gl_minor_version))

        # Shader initialization
        self.cubeShader    = self.initShader("./shaders/pass")
        self.outlineShader = self.initShader("./shaders/outline")
        self.compassShader = self.initShader("./shaders/compass")
        self.screenShader  = self.initShader("./shaders/screen")

        self.compass.initGraphics()
        self.screenQuad.initGraphics()

        self.initFramebuffer(self.size.width, self.size.height)

    def setOutlineMode(self, drawOutlines, outlineWidth = 1.0, outlineDistance = 1000.0):
        self.drawOutlines = drawOutlines
        self.outlineWidth = outlineWidth
        self.outlineDistance = outlineDistance

    def DrawScene(self, backgroundColor, cameraPosition):
        # if self.newPartitionList:
        #     for offset in self.newPartitionList:
        #         self.partitions[offset].initGraphics()
        #     self.newPartitionList = []

        if self.newBufferCount > 0:
            for i in range(self.newBufferCount):
                self.tileBuffers[-(i + 1)].initGraphics()
            self.newBufferCount = 0


        # RENDER TO FRAMEBUFFER
        #======================================================================================
        glBindFramebuffer(GL_FRAMEBUFFER, self.framebuffer)

        glEnable(GL_CULL_FACE)
        glEnable(GL_DEPTH_TEST)
        glDepthFunc(GL_LEQUAL)

        glClear(GL_DEPTH_BUFFER_BIT)
        glClearBufferfv(GL_COLOR, 0, backgroundColor)
        glClearBufferiv(GL_COLOR, 1, [0, 0, 0, -1])

        # DRAW PARTITIONS
        # for partition in self.partitions.itervalues():

        #     glUseProgram(self.cubeShader)

        #     projMatLoc = glGetUniformLocation(self.cubeShader, "u_ProjectionMatrix")
        #     viewMatLoc = glGetUniformLocation(self.cubeShader, "u_ViewMatrix")
        #     glUniformMatrix4fv(projMatLoc, 1, GL_FALSE, self.projectionMatrix)
        #     glUniformMatrix4fv(viewMatLoc, 1, GL_FALSE, self.viewMatrix)
        #     partition.DrawTiles()

        #     if self.drawOutlines and np.linalg.norm(partition.center - cameraPosition) < self.outlineDistance:
        #         glUseProgram(self.outlineShader)
        #         projMatLoc = glGetUniformLocation(self.outlineShader, "u_ProjectionMatrix")
        #         viewMatLoc = glGetUniformLocation(self.outlineShader, "u_ViewMatrix")
        #         glUniformMatrix4fv(projMatLoc, 1, GL_FALSE, self.projectionMatrix)
        #         glUniformMatrix4fv(viewMatLoc, 1, GL_FALSE, self.viewMatrix)
        #         partition.DrawOutlines(self.outlineWidth)

        for tileBuffer in self.tileBuffers:

            glUseProgram(self.cubeShader)

            projMatLoc = glGetUniformLocation(self.cubeShader, "u_ProjectionMatrix")
            viewMatLoc = glGetUniformLocation(self.cubeShader, "u_ViewMatrix")
            glUniformMatrix4fv(projMatLoc, 1, GL_FALSE, self.projectionMatrix)
            glUniformMatrix4fv(viewMatLoc, 1, GL_FALSE, self.viewMatrix)
            tileBuffer.DrawTiles()

            if self.drawOutlines:
                glUseProgram(self.outlineShader)
                projMatLoc = glGetUniformLocation(self.outlineShader, "u_ProjectionMatrix")
                viewMatLoc = glGetUniformLocation(self.outlineShader, "u_ViewMatrix")
                glUniformMatrix4fv(projMatLoc, 1, GL_FALSE, self.projectionMatrix)
                glUniformMatrix4fv(viewMatLoc, 1, GL_FALSE, self.viewMatrix)
                tileBuffer.DrawOutlines(self.outlineWidth)

        # DRAW COMPASS
        glDisable(GL_DEPTH_TEST)

        glUseProgram(self.compassShader)

        screenLoc = glGetUniformLocation(self.compassShader, "u_ScreenSize")
        offsetLoc = glGetUniformLocation(self.compassShader, "u_Offset")
        rotationLoc = glGetUniformLocation(self.compassShader, "u_ViewMatrix")
        glUniform2i(screenLoc, self.size.width, self.size.height)
        glUniform3f(offsetLoc, -0.9, -0.85, 0.0)
        glUniformMatrix4fv(rotationLoc, 1, GL_FALSE, self.viewMatrix)
        self.compass.Draw()

        glBindFramebuffer(GL_FRAMEBUFFER, 0)

        # RENDER TO SCREEN
        #======================================================================================
        glDisable(GL_DEPTH_TEST)
        glDepthFunc(GL_LEQUAL)
        glClear(GL_COLOR_BUFFER_BIT)

        glUseProgram(self.screenShader)

        glActiveTexture(GL_TEXTURE0)
        glBindTexture(GL_TEXTURE_2D, self.screenTex)
        glUniform1i(glGetUniformLocation(self.screenShader, "u_ScreenTexture"), 0)

        self.screenQuad.Draw()

        glUseProgram(0)
